/*
    This is a romfs file system driver for Windows NT and Windows 2000,
    pre-release 2000-07-11.
    Copyright (C) 2000 Bo Branten
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ntifs.h"
#include "fsd.h"
#include "rom_fs.h"
#include "border.h"

ULONG
htonl (
    IN ULONG x
    )
{
    PUCHAR y = (PUCHAR)&x;
    UCHAR tmp;

    tmp = y[0];
    y[0] = y[3];
    y[3] = tmp;
    tmp = y[1];
    y[1] = y[2];
    y[2] = tmp;

    return x;
}

NTSTATUS
FsdCreate (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    PDEVICE_OBJECT      DeviceObject;
    PIRP                Irp;
	PIO_STACK_LOCATION	io_stack;
	NTSTATUS			status;
    PFSD_FCB            root_fcb;
    PFSD_CCB            root_ccb;
    struct romfs_inode* first_file;

    DeviceObject = IrpContext->DeviceObject;
    Irp = IrpContext->Irp;

	io_stack = IoGetCurrentIrpStackLocation(Irp);

    KdPrint((DRIVER_NAME ": Create\n"));

    if (DeviceObject == FsdGlobalData.DeviceObject)
    {
        KdPrint((DRIVER_NAME ": Create: Create on main device object\n"));

        Irp->IoStatus.Information = FILE_OPENED;
        status = STATUS_SUCCESS;
    }
    else if (io_stack->FileObject->FileName.Length == 0)
    {
        io_stack->FileObject->FsContext = DeviceObject->DeviceExtension;
                
        Irp->IoStatus.Information = FILE_OPENED;
        status = STATUS_SUCCESS;
        KdPrint((DRIVER_NAME ": Create on volume succeded\n"));
/*
        KdPrint((DRIVER_NAME ": Create: Create on volume\n"));

        Irp->IoStatus.Information = 0;
        status = STATUS_OBJECT_NAME_NOT_FOUND;
*/
    }
    else
    {
        UNICODE_STRING  slash_string;
        UNICODE_STRING  file_string;
        
        KdPrint((DRIVER_NAME ": Create: Create: (%S)\n", io_stack->FileObject->FileName.Buffer));
        
        RtlInitUnicodeString(&slash_string, L"\\");
        RtlInitUnicodeString(&file_string, L"\\ntifs.h");

            if (!RtlCompareUnicodeString(
                &io_stack->FileObject->FileName,
                &slash_string,
                FALSE
                ))
            {
                root_fcb = ExAllocatePool(NonPagedPool, sizeof(FSD_FCB));
                RtlZeroMemory(root_fcb, sizeof(FSD_FCB));
                root_ccb = ExAllocatePool(NonPagedPool, sizeof(FSD_CCB));
                RtlZeroMemory(root_ccb, sizeof(FSD_CCB));
                root_fcb->Identifier.Type = FCB;
                root_ccb->Identifier.Type = CCB;
                root_fcb->Identifier.Size = sizeof(FSD_FCB);
                root_ccb->Identifier.Size = sizeof(FSD_CCB);

                root_fcb->Vcb = ((PFSD_VCB)DeviceObject->DeviceExtension);
                
                ExInitializeResourceLite(&(root_fcb->MainResource));
                ExInitializeResourceLite(&(root_fcb->PagingIoResource));

                root_fcb->CommonFCBHeader.Resource = &(root_fcb->MainResource);
                root_fcb->CommonFCBHeader.PagingIoResource = &(root_fcb->PagingIoResource);

                root_fcb->CommonFCBHeader.AllocationSize.QuadPart = 4096;
                root_fcb->CommonFCBHeader.FileSize.QuadPart = 4096;

                root_fcb->CommonFCBHeader.ValidDataLength.LowPart = 0xffffffff;
                root_fcb->CommonFCBHeader.ValidDataLength.HighPart = 0x7fffffff;

                root_fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsPossible;
                
                io_stack->FileObject->FsContext = (void*)root_fcb;
                io_stack->FileObject->FsContext2 = (void*)root_ccb;
                io_stack->FileObject->PrivateCacheMap = NULL;
                io_stack->FileObject->SectionObjectPointer = &(root_fcb->SectionObject);

                root_fcb->FileName.Length = 0;

	            root_fcb->FileName.MaximumLength =
                    io_stack->FileObject->FileName.MaximumLength +
                    sizeof(WCHAR);

	            root_fcb->FileName.Buffer = (PWSTR) ExAllocatePool(
                    NonPagedPool,
                    root_fcb->FileName.MaximumLength
                    );

            	if (!root_fcb->FileName.Buffer)
	            {
		            return STATUS_INSUFFICIENT_RESOURCES;
	            }

	            RtlCopyUnicodeString(
                    &root_fcb->FileName,
                    &io_stack->FileObject->FileName
                    );

                root_fcb->FileAttributes = FILE_ATTRIBUTE_NORMAL|FILE_ATTRIBUTE_DIRECTORY;

                root_fcb->romfs_inode = ExAllocatePool(
                    NonPagedPool,
                    sizeof(struct romfs_inode)
                    );

#define ROMFH_SIZE 16
#define ROMFH_PAD (ROMFH_SIZE-1)
#define ROMFH_MASK (~ROMFH_PAD)

                root_fcb->romfs_inode->next = (ROMFH_SIZE +
                    strnlen(root_fcb->Vcb->romfs_super_block->name, ROMFS_MAXFN)
                    + 1 + ROMFH_PAD) & ROMFH_MASK;

                root_fcb->romfs_inode->next = htonl(root_fcb->romfs_inode->next);

                root_ccb->CurrentByteOffset = root_fcb->romfs_inode->next;
                root_ccb->DirectorySearchPattern.Buffer = 0;

                first_file = (struct romfs_inode*)(((PUCHAR)root_fcb->Vcb->romfs_super_block) +
                    root_fcb->romfs_inode->next);

                Irp->IoStatus.Information = FILE_OPENED;
                status = STATUS_SUCCESS;
                KdPrint((DRIVER_NAME ": Create on \\ succeded\n"));
            }
            else if (!RtlCompareUnicodeString(
                &io_stack->FileObject->FileName,
                &file_string,
                FALSE
                ))
            {
                root_fcb = ExAllocatePool(NonPagedPool, sizeof(FSD_FCB));
                RtlZeroMemory(root_fcb, sizeof(FSD_FCB));
                root_ccb = ExAllocatePool(NonPagedPool, sizeof(FSD_CCB));
                RtlZeroMemory(root_ccb, sizeof(FSD_CCB));
                root_fcb->Identifier.Type = FCB;
                root_ccb->Identifier.Type = CCB;
                root_fcb->Identifier.Size = sizeof(FSD_FCB);
                root_ccb->Identifier.Size = sizeof(FSD_CCB);

                root_fcb->Vcb = ((PFSD_VCB)DeviceObject->DeviceExtension);
                
                ExInitializeResourceLite(&(root_fcb->MainResource));
                ExInitializeResourceLite(&(root_fcb->PagingIoResource));

                root_fcb->CommonFCBHeader.Resource = &(root_fcb->MainResource);
                root_fcb->CommonFCBHeader.PagingIoResource = &(root_fcb->PagingIoResource);

                root_fcb->CommonFCBHeader.AllocationSize.QuadPart = 4096;
                root_fcb->CommonFCBHeader.FileSize.QuadPart = 4096;

                root_fcb->CommonFCBHeader.ValidDataLength.LowPart = 0xffffffff;
                root_fcb->CommonFCBHeader.ValidDataLength.HighPart = 0x7fffffff;

                root_fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsPossible;
                
                io_stack->FileObject->FsContext = (void*)root_fcb;
                io_stack->FileObject->FsContext2 = (void*)root_ccb;
                io_stack->FileObject->PrivateCacheMap = NULL;
                io_stack->FileObject->SectionObjectPointer = &(root_fcb->SectionObject);

                root_fcb->FileName.Length = 0;

	            root_fcb->FileName.MaximumLength =
                    io_stack->FileObject->FileName.MaximumLength +
                    sizeof(WCHAR);

	            root_fcb->FileName.Buffer = (PWSTR) ExAllocatePool(
                    NonPagedPool,
                    root_fcb->FileName.MaximumLength
                    );

            	if (!root_fcb->FileName.Buffer)
	            {
		            return STATUS_INSUFFICIENT_RESOURCES;
	            }

	            RtlCopyUnicodeString(
                    &root_fcb->FileName,
                    &io_stack->FileObject->FileName
                    );

                root_fcb->FileAttributes = FILE_ATTRIBUTE_NORMAL;

                root_fcb->romfs_inode = ExAllocatePool(
                    NonPagedPool,
                    sizeof(struct romfs_inode) + ROMFS_MAXFN
                    );

                FsdReadInodeByIndex(
                    root_fcb->Vcb->TargetDeviceObject,
                    0x60000000,
                    root_fcb->romfs_inode);

                root_fcb->IndexNumber.QuadPart = 0x60;
                root_ccb->CurrentByteOffset = 0;

                Irp->IoStatus.Information = FILE_OPENED;
                status = STATUS_SUCCESS;
                KdPrint((DRIVER_NAME ": Create on \\ntifs.h succeded\n"));
            }
            else
            {
                Irp->IoStatus.Information = 0;
                status = STATUS_OBJECT_NAME_NOT_FOUND;
                KdPrint((DRIVER_NAME ": Create denied\n"));
            }
    }

    IrpContext->Irp->IoStatus.Status = status;

    FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);

    FsdFreeIrpContext(IrpContext);

    return status;
}
#if 0
{
    NTSTATUS Status;

    KdPrint((DRIVER_NAME ": FsdCreate\n"));

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

    ASSERT(IrpContext->DeviceObject);
    ASSERT(IrpContext->Irp);

    // just testing the queue
    if (!IrpContext->IsSynchronous)
    {
        KdPrint((DRIVER_NAME ": FsdCreate: Queueing the request\n"));

        FsdQueueRequest(IrpContext);

        return STATUS_PENDING;
    }

    if (IrpContext->DeviceObject == FsdGlobalData.DeviceObject)
    {
        KdPrint((DRIVER_NAME ": FsdCreate: Create on main device object\n"));

        IrpContext->Irp->IoStatus.Information = FILE_OPENED;
        Status = STATUS_SUCCESS;
    }
    else
    {
        // TODO: The rest of the implementation of FsdCreate

        KdPrint((DRIVER_NAME ": FsdCreate: Unimplemented create request\n"));

        IrpContext->Irp->IoStatus.Information = 0;
        Status = STATUS_OBJECT_NAME_NOT_FOUND;
    }

    IrpContext->Irp->IoStatus.Status = Status;

    FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);

    FsdFreeIrpContext(IrpContext);

    return Status;
}
#endif
