/*
    This is a romfs file system driver for Windows NT and Windows 2000,
    pre-release 2000-07-11.
    Copyright (C) 2000 Bo Branten
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ntifs.h"
#include "fsd.h"

#if DBG

static PCHAR IrpMjStrings[] = {
    "IRP_MJ_CREATE",
    "IRP_MJ_CREATE_NAMED_PIPE",
    "IRP_MJ_CLOSE",
    "IRP_MJ_READ",
    "IRP_MJ_WRITE",
    "IRP_MJ_QUERY_INFORMATION",
    "IRP_MJ_SET_INFORMATION",
    "IRP_MJ_QUERY_EA",
    "IRP_MJ_SET_EA",
    "IRP_MJ_FLUSH_BUFFERS",
    "IRP_MJ_QUERY_VOLUME_INFORMATION",
    "IRP_MJ_SET_VOLUME_INFORMATION",
    "IRP_MJ_DIRECTORY_CONTROL",
    "IRP_MJ_FILE_SYSTEM_CONTROL",
    "IRP_MJ_DEVICE_CONTROL",
    "IRP_MJ_INTERNAL_DEVICE_CONTROL",
    "IRP_MJ_SHUTDOWN",
    "IRP_MJ_LOCK_CONTROL",
    "IRP_MJ_CLEANUP",
    "IRP_MJ_CREATE_MAILSLOT",
    "IRP_MJ_QUERY_SECURITY",
    "IRP_MJ_SET_SECURITY",
    "IRP_MJ_POWER",
    "IRP_MJ_SYSTEM_CONTROL",
    "IRP_MJ_DEVICE_CHANGE",
    "IRP_MJ_QUERY_QUOTA",
    "IRP_MJ_SET_QUOTA",
    "IRP_MJ_PNP"
};

VOID
FsdDbgPrintCall (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    )
{
    PIO_STACK_LOCATION  IoStackLocation;
    PFSD_FCB            Fcb;
    PWCHAR              FileName = L"Unknown";

    IoStackLocation = IoGetCurrentIrpStackLocation(Irp);

    if (DeviceObject == FsdGlobalData.DeviceObject)
    {
        FileName = DEVICE_NAME;
    }
    else if (IoStackLocation->MajorFunction == IRP_MJ_CREATE &&
             IoStackLocation->FileObject)
    {
        if (IoStackLocation->FileObject->FileName.Length == 0)
        {
            FileName = L"Volume";
        }
        else if (IoStackLocation->FileObject->FileName.Buffer)
        {
            FileName = IoStackLocation->FileObject->FileName.Buffer;
        }
    }
    else if (IoStackLocation->FileObject &&
             IoStackLocation->FileObject->FsContext)
    {
        Fcb = (PFSD_FCB) IoStackLocation->FileObject->FsContext;

        if (Fcb->Identifier.Type == FCB &&
            Fcb->Identifier.Size == sizeof(FSD_FCB) &&
            Fcb->FileName.Buffer)
        {
            FileName = Fcb->FileName.Buffer;
        }
    }

    DbgPrint(
        DRIVER_NAME ": %s %S\n",
        IrpMjStrings[IoStackLocation->MajorFunction],
        FileName
        );
}

VOID
FsdDbgPrintComplete (
    IN PIRP     Irp,
    IN CCHAR    PriorityBoost
    )
{
    PIO_STACK_LOCATION  IoStackLocation;
    PFSD_FCB            Fcb;
    PWCHAR              FileName = L"Unknown";
    PCHAR               StatusString = NULL;

    if (!NT_SUCCESS(Irp->IoStatus.Status))
    {
    IoStackLocation = IoGetCurrentIrpStackLocation(Irp);

    if (IoStackLocation->DeviceObject == FsdGlobalData.DeviceObject)
    {
        FileName = DEVICE_NAME;
    }
    else if (IoStackLocation->MajorFunction == IRP_MJ_CREATE &&
             IoStackLocation->FileObject)
    {
        if (IoStackLocation->FileObject->FileName.Length == 0)
        {
            FileName = L"Volume";
        }
        else if (IoStackLocation->FileObject->FileName.Buffer)
        {
            FileName = IoStackLocation->FileObject->FileName.Buffer;
        }
    }
    else if (IoStackLocation->FileObject &&
             IoStackLocation->FileObject->FsContext)
    {
        Fcb = (PFSD_FCB) IoStackLocation->FileObject->FsContext;

        if (Fcb->Identifier.Type == FCB &&
            Fcb->Identifier.Size == sizeof(FSD_FCB) &&
            Fcb->FileName.Buffer)
        {
            FileName = Fcb->FileName.Buffer;
        }
    }

        switch (Irp->IoStatus.Status)
        {
        case STATUS_BUFFER_OVERFLOW:
            StatusString = "STATUS_BUFFER_OVERFLOW";
            break;

        case STATUS_BUFFER_TOO_SMALL:
            StatusString = "STATUS_BUFFER_TOO_SMALL";
            break;

        case STATUS_DRIVER_INTERNAL_ERROR:
            StatusString = "STATUS_DRIVER_INTERNAL_ERROR";
            break;

        case STATUS_END_OF_FILE:
            StatusString = "STATUS_END_OF_FILE";
            break;

        case STATUS_INSUFFICIENT_RESOURCES:
            StatusString = "STATUS_INSUFFICIENT_RESOURCES";
            break;

        case STATUS_INVALID_PARAMETER:
            StatusString = "STATUS_INVALID_PARAMETER";
            break;

        case STATUS_INVALID_USER_BUFFER:
            StatusString = "STATUS_INVALID_USER_BUFFER";
            break;

        case STATUS_MEDIA_WRITE_PROTECTED:
            StatusString = "STATUS_MEDIA_WRITE_PROTECTED";
            break;

        case STATUS_NO_MORE_FILES:
            StatusString = "STATUS_NO_MORE_FILES";
            break;

        case STATUS_NO_SUCH_FILE:
            StatusString = "STATUS_NO_SUCH_FILE";
            break;

        case STATUS_NOT_IMPLEMENTED:
            StatusString = "STATUS_NOT_IMPLEMENTED";
            break;

        case STATUS_OBJECT_NAME_INVALID:
            StatusString = "STATUS_OBJECT_NAME_INVALID";
            break;

        case STATUS_OBJECT_NAME_NOT_FOUND:
            StatusString = "STATUS_OBJECT_NAME_NOT_FOUND";
            break;

        case STATUS_UNRECOGNIZED_VOLUME:
            StatusString = "STATUS_UNRECOGNIZED_VOLUME";
            break;

        case STATUS_UNSUCCESSFUL:
            StatusString = "STATUS_UNSUCCESSFUL";
            break;
        }

        if (StatusString)
        {
            DbgPrint(
                DRIVER_NAME ": %s %s %S\n",
                IrpMjStrings[IoStackLocation->MajorFunction],
                StatusString,
                FileName
                );
        }
        else
        {
            DbgPrint(
                DRIVER_NAME ": %s STATUS=0X%X %S\n",
                IrpMjStrings[IoStackLocation->MajorFunction],
                Irp->IoStatus.Status,
                FileName
                );
        }
    }
}

#endif // DBG
